#include <windows.h>
#include <limits.h>
#include <stdio.h>
#include <conio.h>

#define DEV_NAME	"\\\\.\\PWLVIRTUALJOY"



enum motion_kind {
    k_none, k_axis, k_buttons, k_hat, k_exit
};

#define C_button 0
#define C_axis 1
#define C_hat 2

struct ioctl_data {
    int	command;
    int index;
    int	value;
};

int callDriver(HANDLE hDriver, int cmd, int id, int val)
{
/*
DeviceIoControl
The DeviceIoControl function sends a control code directly to a specified device driver, causing the corresponding device to perform the corresponding operation. 

BOOL DeviceIoControl(
  HANDLE hDevice,              // handle to device
  DWORD dwIoControlCode,       // operation
  LPVOID lpInBuffer,           // input data buffer
  DWORD nInBufferSize,         // size of input data buffer
  LPVOID lpOutBuffer,          // output data buffer
  DWORD nOutBufferSize,        // size of output data buffer
  LPDWORD lpBytesReturned,     // byte count
  LPOVERLAPPED lpOverlapped    // overlapped information
);
*/
    struct ioctl_data data;
    DWORD bytesRet;

    data.command = cmd;
    data.index = id;
    data.value = val;
    return DeviceIoControl(hDriver, 0, &data, sizeof(data), /*NULL*/&bytesRet, 4, &bytesRet, NULL);
}

void main()
{
    enum motion_kind kind = k_none;
    HANDLE hDriver;
    int id, step, val;
    int res = 1;

printf("Opening: %s\n",DEV_NAME);
    if ((hDriver = CreateFile(DEV_NAME,
		GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL, NULL)) != ((HANDLE)-1)) {
	printf("Device opened succesfully with handle 0x%04x\n", hDriver);
    } else {
	char buf[1024*10];
	printf("Can't open device\n");
	FormatMessage(/*FORMAT_MESSAGE_ALLOCATE_BUFFER*/FORMAT_MESSAGE_FROM_SYSTEM,0,GetLastError(),0,buf,1024*10,NULL);
	printf("err: %s (%d)\n", buf, GetLastError());
    }

    while(k_exit != kind) {
	switch (kind) {
	case k_axis:
	    for (id = 0; id < 16 && res; id++) {
		step = 128;
		val = -SHRT_MAX;
		for (; val < SHRT_MAX; val+=step) {
		    res = callDriver(hDriver, C_axis, id, val);
		    if (res == 0) {
			break;
		    }
		    Sleep(1);
		    printf("<%d,%d>       \r",id,val);
		    if (val > SHRT_MAX) {
			val = SHRT_MAX;
			step = -step;
		    }
		    if (val < -SHRT_MAX) {
			val = -SHRT_MAX;
			step = -step;
		    }

		}
		printf("\n");
	    }
	    printf("finished with axes\n", id);
	    kind = k_hat;
	    break;
	case k_buttons:
	    res = 1;
	    for (id = 0; id < INT_MAX && res; id++) {
		for (val = 0; val < 3; val++) {
		    res = callDriver(hDriver, C_button, id, val & 1);
		    if (res == 0)
			break;
		    printf("<%d,%d>\r",id,val);
		    Sleep(200);
		}

	    }
	    kind = k_exit;
	    break;
	case k_hat:
	    step = 1;
	    val = 0;
	    for (; val <= 360; val+=step) {
		    res = callDriver(hDriver, C_hat, 0, val);
		    if (res == 0) {
			break;
		    }
		    Sleep(1);
		    printf("<%d,%d>       \r",id,val);
		    if (val > SHRT_MAX) {
			val = SHRT_MAX;
			step = -step;
		    }
		    if (val < -SHRT_MAX) {
			val = -SHRT_MAX;
			step = -step;
		    }

		}
	    printf("finished with hat\n", id);
	    kind = k_buttons;
	    break;
	case k_none:
	    kind = k_axis;
	    continue;
	default:
	    break;
	}
    }
    printf("finished\n");
    getch();
}
